#!/usr/bin/env python3
# coding: utf-8

"""socks5server.py: simple socks5 server"""
__version__   = "2019-05-02"
__license__   = "MIT"
__author__    = "ed <irc.rizon.net>"
__credits__   = ["Artem Golubin", "stackoverflow.com"]

# grab the latest ver:
# -- https://ocv.me/dev/?socks5server.py
#
# heavily based on work by Artem Golubin (2018-04-19, MIT-Licensed):
# -- https://rushter.com/blog/python-socks-server/
# -- https://github.com/rushter/socks5


import os
import sys
import logging
import select
import socket
import struct
from socketserver import ThreadingMixIn, TCPServer, StreamRequestHandler


logging.basicConfig(
	level = logging.DEBUG,
	format = '\033[36m%(asctime)s \033[33m%(levelname)5s \033[0m%(message)s'
)

SOCKS_VERSION = 5

DEST_IPV4     = 1
DEST_DOMAIN   = 3
DEST_IPV6     = 4

CMD_STREAM    = 1


class ThreadingTCPServer(ThreadingMixIn, TCPServer):
	pass


class SocksProxy(StreamRequestHandler):
	def read_string(self):
		strlen = self.connection.recv(1)[0]
		buf = self.connection.recv(strlen)
		assert len(buf) == strlen
		return buf.decode('utf-8')
	
	
	def negotiate_auth_methods(self):
		# read SOCKS version and authentication methods from client
		version, num_methods = struct.unpack("!BB",
			self.connection.recv(2))

		if version != SOCKS_VERSION:
			logging.info('Rejecting client; invalid SOCKS version: {0}'.format(version))
			return False

		if num_methods <= 0:
			logging.info('Rejecting client; no auth methods suggested')
			return False
		
		methods = self.connection.recv(num_methods)
		
		str_methods = ','.join(str(int(x)) for x in methods)
		logging.debug('Client offers auth methods [{0}]'.format(str_methods))
		
		if 2 in set(methods) and self.server.password is not None:
			# reply that we accept the offered user:pass authentication
			self.connection.sendall(struct.pack("!BB", SOCKS_VERSION, 2))
		
		elif 0 in set(methods) and self.server.password is None:
			# reply that we accept unauthenticated connections
			self.connection.sendall(struct.pack("!BB", SOCKS_VERSION, 0))
		
		else:
			logging.info('We support auth methods "2" (user:pass) and "0" (allow-all); client offers [{0}]'.format(str_methods))
			return False
		
		return True
	
	
	def verify_credentials(self):
		if self.server.password is None:
			# authentication disabled
			return True
		
		version = self.connection.recv(1)[0]
		if version != 1:
			logging.info('Expected user:pass auth ver.1, got {0}'.format(version))
			return False

		username = self.read_string()
		password = self.read_string()

		# reply 0 on success, >0 on error
		if username == self.server.username \
		and password == self.server.password:
			response = struct.pack("!BB", version, 0)
		else:
			response = struct.pack("!BB", version, 0xFF)
			logger.info('Failed login attempt: {0} / {1}'.format(
				username, password))
		
		self.connection.sendall(response)
		return response[1] == 0


	def generate_success_response(self, ip_ver, addr, port):
		if ip_ver == socket.AF_INET:
			address_type = DEST_IPV4
		else:
			address_type = DEST_IPV6
		
		addr = socket.inet_pton(ip_ver, addr)
		
		return \
			struct.pack("!BBBB", SOCKS_VERSION, 0, 0, address_type) + \
			addr + \
			struct.pack('!H', port)
		
		
	def generate_error_response(self, address_type, error_number):
		return struct.pack("!BBBBIH",
			SOCKS_VERSION, error_number, 0, address_type, 0, 0)


	def exchange_loop(self, client, remote):
		while True:
			# wait until client or remote is available for read
			r, w, e = select.select([client, remote], [], [])

			if client in r:
				data = client.recv(4096)
				if remote.send(data) <= 0:
					break

			if remote in r:
				data = remote.recv(4096)
				if client.send(data) <= 0:
					break


	def end(self):
		client_addr, client_port = self.client_address
		logging.info('Terminating {0}/{1}'.format(
			client_addr, client_port))
		
		self.server.close_request(self.request)


	def handle(self):
		client_addr, client_port = self.client_address
		logging.info('Accepting {0}/{1}'.format(
			client_addr, client_port))

		if not self.negotiate_auth_methods() \
		or not self.verify_credentials():
			self.end()
			return

		version, cmd, _, address_type = \
			struct.unpack("!BBBB", self.connection.recv(4))
		
		assert version == SOCKS_VERSION
		
		if address_type == DEST_IPV4:
			logging.debug('Destination is an IPv4 address')
			address = socket.inet_ntoa(self.connection.recv(4))
			ip_ver = socket.AF_INET
		
		elif address_type == DEST_DOMAIN:
			logging.debug('Destination is a domain name')
			address = self.read_string()
			ip_ver = socket.AF_INET  # bad assumption, TODO
		
		elif address_type == DEST_IPV6:
			logging.debug('Destination is an IPv6 address')
			address = socket.inet_ntop(socket.AF_INET6, self.connection.recv(16))
			ip_ver = socket.AF_INET6
		
		else:
			logging.info('Unsupported destination type: {0}'.format(address_type))
			self.end()
			return

		port = struct.unpack('!H', self.connection.recv(2))[0]
		logging.debug('Destination: {0}:{1}'.format(address, port))

		try:
			if cmd == CMD_STREAM:  # tcp/ip stream connection
				remote = socket.socket(ip_ver, socket.SOCK_STREAM)
				remote.connect((address, port))
				bind_address = remote.getsockname()
				logging.info('Connected to {0}/{1}'.format(*bind_address))
			else:
				logging.info('Unsupported command: {0}'.format(cmd))
				self.end()

			addr, port = bind_address[:2]
			reply = self.generate_success_response(ip_ver, addr, port)

		except Exception as err:
			logging.error(err)
			# error 5 = connection refused
			reply = self.generate_error_response(address_type, 5)

		self.connection.sendall(reply)

		# establish data exchange
		if reply[1] == 0 and cmd == CMD_STREAM:
			self.exchange_loop(self.connection, remote)

		self.end()


def main():
	example = 'examples:\n  \033[32mUP="ed:hunter2" {0} 0.0.0.0 43214\n  UP="-" {0} 127.0.0.1 43214'.format(sys.argv[0])
	
	username = None
	password = None
	try:
		env = os.environ['UP']
		if env != '-':
			username, password = env.split(':', 1)
			assert len(username) > 1
			assert len(password) > 1
	except:
		logging.error('\n  need username:password in the environment variable UP,\n  or set UP="-" to disable authentication entirely')
		logging.info(example)
		return
	
	try:
		endpoint = sys.argv[1]
		port = int(sys.argv[2])
		assert len(endpoint) >= len('0.0.0.0')
	except:
		logging.error('need arg 1:  ip address to bind')
		logging.error('need arg 2:  port to listen on')
		logging.info(example)
		return
	
	if endpoint == '127.0.0.1':
		preface = "only accepting localhost connections "
		if password is None:
			logging.info(preface + "(without authentication)")
		else:
			logging.info(preface)
	else:
		preface = "accepting external connections with "
		if password is None:
			logging.warn(preface + "ZERO authentication... sure hope you know what you're doing")
		else:
			logging.info(preface + "username/password authentication")
	
	TCPServer.allow_reuse_address = True
	
	server = ThreadingTCPServer((endpoint, port), SocksProxy)
	server.username = username
	server.password = password
	logging.info('Listening on {0}/{1}'.format(endpoint, port))
	server.serve_forever()


if __name__ == '__main__':
	main()


# UP=ed:hunter2 nice stdbuf -oL ./socks5server.py 0.0.0.0 43214 | tee socks5.$(date +%Y-%m%d-%H%M%S)
# curl -v --socks5 127.0.0.1:43214 -U ed:hunter2 http://ip.ocv.me/
# nslookup -query=AAAA google.com
# curl -v --socks5 127.0.0.1:43214 -U ed:hunter2 https://'[2a00:1450:400f:80a::200e]'/

# untrusted socks5:43214 is «glitch», main/trusted machine is «orz», untrusted storage machine is «crate»
# orz> ssh -t -L 43215:0.0.0.0:43214 glitch bash
# orz> ssh -t -R :43215:127.0.0.1:43215 crate
# socks5 is now accessible on crate through 127.0.0.1:43215

